from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pygame
from pygame import surface, time

from ..colors import LIGHT_GRAY, RED, WHITE, YELLOW
from ..font_utils import load_font, render_text_surface
from ..gameplay_constants import (
    CAR_HINT_DELAY_MS_DEFAULT,
    SURVIVAL_TIME_ACCEL_SUBSTEPS,
    SURVIVAL_TIME_ACCEL_MAX_SUBSTEP,
)
from ..gameplay import (
    MapGenerationError,
    apply_passenger_speed_penalty,
    check_interactions,
    cleanup_survivor_messages,
    generate_level_from_blueprint,
    initialize_game_state,
    maintain_waiting_car_supply,
    nearest_waiting_car,
    schedule_timed_message,
    place_flashlights,
    place_fuel_can,
    place_shoes,
    process_player_input,
    setup_player_and_cars,
    spawn_initial_patrol_bots,
    spawn_initial_zombies,
    spawn_survivors,
    sync_ambient_palette_with_flashlights,
    update_entities,
    update_footprints,
    update_endurance_timer,
)
from ..gameplay.state import frames_to_ms, ms_to_frames
from ..gameplay.constants import (
    INTRO_MESSAGE_DISPLAY_FRAMES,
    LAYER_ITEMS,
)
from ..input_utils import (
    CONTROLLER_BUTTON_DOWN,
    CONTROLLER_DEVICE_ADDED,
    CONTROLLER_DEVICE_REMOVED,
    init_first_controller,
    init_first_joystick,
    is_accel_active,
    is_select_event,
    is_start_event,
    read_gamepad_move,
)
from ..gameplay.spawn import _alive_waiting_cars
from ..world_grid import build_wall_index
from ..entities.walls import consume_wall_index_dirty
from ..localization import get_font_settings, translate as tr
from ..models import Stage
from ..render import (
    draw,
    draw_debug_overview,
    draw_pause_overlay,
    prewarm_fog_overlays,
    blit_message_wrapped,
)
from ..render_constants import (
    GAMEPLAY_FONT_SIZE,
    TIMED_MESSAGE_LEFT_X,
    TIMED_MESSAGE_TOP_Y,
)
from ..rng import generate_seed, seed_rng
from ..progress import record_stage_clear
from ..screens import ScreenID, ScreenTransition
from ..windowing import nudge_window_scale, present, sync_window_size, toggle_fullscreen

if TYPE_CHECKING:
    from ..render import RenderAssets


_SHARED_FOG_CACHE: dict[str, Any] | None = None


def gameplay_screen(
    screen: surface.Surface,
    clock: time.Clock,
    config: dict[str, Any],
    fps: int,
    stage: Stage,
    *,
    show_pause_overlay: bool,
    seed: int | None,
    render_assets: "RenderAssets",
    debug_mode: bool = False,
    show_fps: bool = False,
    profiler: "object | None" = None,
    profiler_output: Path | None = None,
) -> ScreenTransition:
    """Main gameplay loop that returns the next screen transition."""

    screen_width = screen.get_width()
    screen_height = screen.get_height()
    mouse_hidden = False
    use_busy_loop = os.environ.get("ZOMBIE_ESCAPE_BUSY_LOOP") == "1"
    profiling_active = False

    def _dump_profile() -> None:
        nonlocal profiling_active
        if profiler is None or profiler_output is None:
            return
        try:
            import pstats
        except Exception:
            return
        if profiling_active:
            profiler.disable()
            profiling_active = False
        output_path = profiler_output
        profiler.dump_stats(output_path)
        summary_path = output_path.with_suffix(".txt")
        with summary_path.open("w", encoding="utf-8") as handle:
            stats = pstats.Stats(profiler, stream=handle).sort_stats("tottime")
            stats.print_stats(50)
        print(f"Profile saved to {output_path} and {summary_path}")

    def _set_mouse_hidden(hidden: bool) -> None:
        nonlocal mouse_hidden
        if mouse_hidden == hidden:
            return
        pygame.mouse.set_visible(not hidden)
        mouse_hidden = hidden

    def _finalize(transition: ScreenTransition) -> ScreenTransition:
        _set_mouse_hidden(False)
        if profiling_active:
            _dump_profile()
        return transition

    def _show_loading_still() -> None:
        screen.fill((0, 0, 0))
        if stage.intro_key:
            intro_text = tr(stage.intro_key)
            font_settings = get_font_settings()
            font_size = font_settings.scaled_size(GAMEPLAY_FONT_SIZE * 2)
            font = load_font(font_settings.resource, font_size)
            line_height = int(
                round(font.get_linesize() * font_settings.line_height_scale)
            )
            x = TIMED_MESSAGE_LEFT_X
            y = TIMED_MESSAGE_TOP_Y
            for line in intro_text.splitlines():
                if not line:
                    y += line_height
                    continue
                surface = font.render(line, False, WHITE)
                screen.blit(surface, (x, y))
                y += line_height
        present(screen)
        pygame.event.pump()

    seed_value = seed if seed is not None else generate_seed()
    applied_seed = seed_rng(seed_value)

    loading_started_ms = pygame.time.get_ticks()
    _show_loading_still()

    game_data = initialize_game_state(config, stage)
    game_data.state.seed = applied_seed
    game_data.state.debug_mode = debug_mode
    game_data.state.show_fps = show_fps
    # if debug_mode and stage.endurance_stage:
    #     goal_ms = max(0, stage.endurance_goal_ms)
    #     if goal_ms > 0:
    #         remaining = 3 * 60 * 1000  # 3 minutes in ms
    #         game_data.state.endurance_elapsed_ms = max(0, goal_ms - remaining)
    #         game_data.state.dawn_ready = False
    #         game_data.state.dawn_prompt_at = None
    #         game_data.state.dawn_carbonized = False
    global _SHARED_FOG_CACHE
    if _SHARED_FOG_CACHE is None:
        prewarm_fog_overlays(
            game_data.fog,
            render_assets,
            stage=stage,
        )
        _SHARED_FOG_CACHE = game_data.fog
    else:
        game_data.fog = _SHARED_FOG_CACHE
    if stage.intro_key and game_data.state.timed_message:
        loading_elapsed_ms = max(0, pygame.time.get_ticks() - loading_started_ms)
        remaining_ms = max(
            0, frames_to_ms(INTRO_MESSAGE_DISPLAY_FRAMES) - loading_elapsed_ms
        )
        schedule_timed_message(
            game_data.state,
            tr(stage.intro_key),
            duration_frames=max(0, ms_to_frames(remaining_ms)),
            clear_on_input=True,
            color=LIGHT_GRAY,
            align="left",
            now_ms=game_data.state.clock.elapsed_ms,
        )
    paused_manual = False
    paused_focus = False
    ignore_focus_loss_until = 0
    debug_overview = False
    controller = init_first_controller()
    joystick = init_first_joystick() if controller is None else None
    _set_mouse_hidden(pygame.mouse.get_focused())

    try:
        layout, layout_data, wall_group, all_sprites, blueprint = (
            generate_level_from_blueprint(
                stage,
                config,
                seed=game_data.state.seed,
                ambient_palette_key=game_data.state.ambient_palette_key,
            )
        )
        game_data.layout = layout
        game_data.blueprint = blueprint
        game_data.groups.wall_group = wall_group
        game_data.groups.all_sprites = all_sprites
        game_data.wall_index_dirty = True
    except MapGenerationError:
        # If generation fails after retries, show error and back to title
        screen.fill((0, 0, 0))
        blit_message_wrapped(
            screen,
            tr("errors.map_generation_failed"),
            16,
            RED,
            (screen_width // 2, screen_height // 2),
            max_width=screen_width - 40,
        )
        present(screen)
        pygame.time.delay(3000)
        return _finalize(ScreenTransition(ScreenID.TITLE))

    sync_ambient_palette_with_flashlights(game_data, force=True)
    initial_waiting = max(0, stage.waiting_car_target_count)
    player, waiting_cars = setup_player_and_cars(
        game_data, layout_data, car_count=initial_waiting
    )
    game_data.player = player
    game_data.waiting_cars = waiting_cars
    game_data.car = None
    # Only top up if initial placement spawned fewer than the intended baseline (shouldn't happen)
    maintain_waiting_car_supply(game_data, minimum=stage.waiting_car_target_count)
    apply_passenger_speed_penalty(game_data)

    spawn_survivors(game_data, layout_data)

    occupied_centers: set[tuple[int, int]] = set()
    cell_size = game_data.cell_size
    if stage.requires_fuel:
        fuel_spawn_count = stage.fuel_spawn_count
        if stage.endurance_stage:
            fuel_spawn_count = 0
        fuel_can = place_fuel_can(
            layout_data["fuel_cells"],
            cell_size,
            player,
            cars=game_data.waiting_cars,
            reserved_centers=occupied_centers,
            count=fuel_spawn_count,
        )
        if fuel_can:
            game_data.fuel = fuel_can
            game_data.groups.all_sprites.add(fuel_can, layer=LAYER_ITEMS)
            occupied_centers.add(fuel_can.rect.center)
    flashlight_count = stage.initial_flashlight_count
    flashlights = place_flashlights(
        layout_data["flashlight_cells"],
        cell_size,
        player,
        cars=game_data.waiting_cars,
        reserved_centers=occupied_centers,
        count=max(0, flashlight_count),
    )
    game_data.flashlights = flashlights
    game_data.groups.all_sprites.add(flashlights, layer=LAYER_ITEMS)
    for flashlight in flashlights:
        occupied_centers.add(flashlight.rect.center)

    shoes_count = stage.initial_shoes_count
    shoes_list = place_shoes(
        layout_data["shoes_cells"],
        cell_size,
        player,
        cars=game_data.waiting_cars,
        reserved_centers=occupied_centers,
        count=max(0, shoes_count),
    )
    game_data.shoes = shoes_list
    game_data.groups.all_sprites.add(shoes_list, layer=LAYER_ITEMS)

    spawn_initial_zombies(game_data, player, layout_data, config)
    spawn_initial_patrol_bots(game_data, player, layout_data)
    update_footprints(game_data, config)
    level_rect = game_data.layout.field_rect
    overview_surface = pygame.Surface((level_rect.width, level_rect.height))
    while True:
        frame_ms = clock.tick_busy_loop(fps) if use_busy_loop else clock.tick(fps)
        dt = frame_ms / 1000.0
        current_fps = clock.get_fps()
        if game_data.state.game_over or game_data.state.game_won:
            game_data.state.clock.time_scale = 1.0
            game_data.state.clock.tick(frame_ms)
            if game_data.state.game_won:
                record_stage_clear(stage.id)
            if game_data.state.game_over and not game_data.state.game_won:
                if game_data.state.game_over_at is None:
                    game_data.state.game_over_at = game_data.state.clock.elapsed_ms
                if game_data.state.clock.elapsed_ms - game_data.state.game_over_at < 1000:
                    draw(
                        render_assets,
                        screen,
                        game_data,
                        config=config,
                        hint_color=None,
                        fps=current_fps,
                    )
                    present(screen)
                    continue
            return _finalize(
                ScreenTransition(
                    ScreenID.GAME_OVER,
                    stage=stage,
                    game_data=game_data,
                    config=config,
                )
            )

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return _finalize(ScreenTransition(ScreenID.EXIT))
            if event.type in (pygame.WINDOWSIZECHANGED, pygame.VIDEORESIZE):
                sync_window_size(event, game_data=game_data)
                continue
            if event.type == pygame.WINDOWFOCUSLOST:
                now = game_data.state.clock.elapsed_ms
                if now >= ignore_focus_loss_until:
                    paused_focus = True
            if event.type == pygame.WINDOWFOCUSGAINED:
                paused_focus = False
            if event.type == pygame.MOUSEBUTTONDOWN:
                paused_focus = False
            if event.type == pygame.JOYDEVICEADDED or (
                CONTROLLER_DEVICE_ADDED is not None
                and event.type == CONTROLLER_DEVICE_ADDED
            ):
                if controller is None:
                    controller = init_first_controller()
                if controller is None:
                    joystick = init_first_joystick()
            if event.type == pygame.JOYDEVICEREMOVED or (
                CONTROLLER_DEVICE_REMOVED is not None
                and event.type == CONTROLLER_DEVICE_REMOVED
            ):
                if controller and not controller.get_init():
                    controller = None
                if joystick and not joystick.get_init():
                    joystick = None
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_F10:
                    if profiler is not None:
                        if profiling_active:
                            _dump_profile()
                        else:
                            profiler.enable()
                            profiling_active = True
                            print("Profile started (F10 to stop and save).")
                    continue
                if event.key == pygame.K_LEFTBRACKET:
                    nudge_window_scale(0.5, game_data=game_data)
                    continue
                if event.key == pygame.K_RIGHTBRACKET:
                    nudge_window_scale(2.0, game_data=game_data)
                    continue
                if event.key == pygame.K_f:
                    toggle_fullscreen(game_data=game_data)
                    continue
                if event.key == pygame.K_s and (
                    pygame.key.get_mods() & pygame.KMOD_CTRL
                ):
                    state_snapshot = {
                        k: v
                        for k, v in vars(game_data.state).items()
                        if k != "footprints"
                    }
                    print("STATE DEBUG:", state_snapshot)
                    continue
                if debug_mode:
                    if event.key == pygame.K_ESCAPE:
                        return ScreenTransition(ScreenID.TITLE)
                    if event.key == pygame.K_p:
                        paused_manual = not paused_manual
                    if event.key == pygame.K_o:
                        debug_overview = not debug_overview
                    continue
                if paused_manual:
                    if event.key == pygame.K_ESCAPE:
                        return _finalize(ScreenTransition(ScreenID.TITLE))
                    if event.key == pygame.K_p:
                        paused_manual = False
                    continue
                if event.key in (pygame.K_ESCAPE, pygame.K_p):
                    paused_manual = True
                    continue
            if event.type == pygame.JOYBUTTONDOWN or (
                CONTROLLER_BUTTON_DOWN is not None
                and event.type == CONTROLLER_BUTTON_DOWN
            ):
                if debug_mode:
                    if is_select_event(event):
                        return _finalize(ScreenTransition(ScreenID.TITLE))
                    if is_start_event(event):
                        paused_manual = not paused_manual
                    continue
                if paused_manual:
                    if is_select_event(event):
                        return _finalize(ScreenTransition(ScreenID.TITLE))
                    if is_start_event(event):
                        paused_manual = False
                    continue
                if is_select_event(event) or is_start_event(event):
                    paused_manual = True
                    continue

        _set_mouse_hidden(pygame.mouse.get_focused())

        paused = paused_manual or paused_focus
        if paused:
            if debug_overview:
                draw_debug_overview(
                    render_assets,
                    screen,
                    overview_surface,
                    game_data,
                    config,
                    screen_width=screen_width,
                    screen_height=screen_height,
                )
            else:
                draw(
                    render_assets,
                    screen,
                    game_data,
                    config=config,
                    fps=current_fps,
                )
            if show_pause_overlay:
                draw_pause_overlay(screen)
            present(screen)
            continue

        keys = pygame.key.get_pressed()
        accel_allowed = not (game_data.state.game_over or game_data.state.game_won)
        accel_active = accel_allowed and is_accel_active(keys, controller, joystick)
        game_data.state.time_accel_active = accel_active
        substeps = SURVIVAL_TIME_ACCEL_SUBSTEPS if accel_active else 1
        sub_dt = min(dt, SURVIVAL_TIME_ACCEL_MAX_SUBSTEP) if accel_active else dt
        if consume_wall_index_dirty():
            game_data.wall_index_dirty = True
        if game_data.wall_index is None or game_data.wall_index_dirty:
            game_data.wall_index = build_wall_index(
                game_data.groups.wall_group, cell_size=game_data.cell_size
            )
            game_data.wall_index_dirty = False
        wall_index = game_data.wall_index
        for _ in range(substeps):
            player_ref = game_data.player
            if player_ref is None:
                break
            car_ref = game_data.car
            pad_vector = read_gamepad_move(controller, joystick)
            player_dx, player_dy, car_dx, car_dy = process_player_input(
                keys,
                player_ref,
                car_ref,
                shoes_count=game_data.state.shoes_count,
                pad_input=pad_vector,
            )
            if (
                game_data.state.timed_message
                and game_data.state.timed_message.clear_on_input
                and (player_dx or player_dy or car_dx or car_dy)
            ):
                game_data.state.timed_message = None
            update_entities(
                game_data,
                player_dx,
                player_dy,
                car_dx,
                car_dy,
                config,
                wall_index=wall_index,
            )
            update_footprints(game_data, config)
            step_ms = int(sub_dt * 1000)
            if accel_active:
                step_ms = max(1, step_ms)
            time_scale = 4.0 / max(1, substeps) if accel_active else 1.0
            game_data.state.clock.time_scale = time_scale
            step_ms = game_data.state.clock.tick(step_ms)
            update_endurance_timer(game_data, step_ms)
            cleanup_survivor_messages(game_data.state)
            check_interactions(game_data, config)
            if game_data.state.game_over or game_data.state.game_won:
                break

        player_ref = game_data.player
        if player_ref is not None:
            mobile_entities: list[pygame.sprite.Sprite] = []
            if player_ref.alive():
                mobile_entities.append(player_ref)
            car_ref = game_data.car
            if car_ref and car_ref.alive():
                mobile_entities.append(car_ref)
            mobile_entities.extend(
                [zombie for zombie in game_data.groups.zombie_group if zombie.alive()]
            )
            mobile_entities.extend(
                [
                    survivor
                    for survivor in game_data.groups.survivor_group
                    if survivor.alive()
                ]
            )
            mobile_entities.extend(
                [
                    bot
                    for bot in game_data.groups.patrol_bot_group
                    if bot.alive()
                ]
            )
            game_data.state.spatial_index.rebuild(mobile_entities)

        player = game_data.player
        if player is None:
            raise ValueError("Player missing from game data")

        if debug_overview:
            draw_debug_overview(
                render_assets,
                screen,
                overview_surface,
                game_data,
                config,
                screen_width=screen_width,
                screen_height=screen_height,
            )
            present(screen)
            continue

        car_hint_conf = config.get("car_hint", {})
        hint_delay = car_hint_conf.get("delay_ms", CAR_HINT_DELAY_MS_DEFAULT)
        elapsed_ms = game_data.state.clock.elapsed_ms
        has_fuel = game_data.state.has_fuel
        hint_enabled = car_hint_conf.get("enabled", True) and not stage.endurance_stage
        hint_target = None
        hint_color = YELLOW
        hint_expires_at = game_data.state.hint_expires_at
        hint_target_type = game_data.state.hint_target_type

        active_car = game_data.car if game_data.car and game_data.car.alive() else None
        if hint_enabled:
            if not has_fuel and game_data.fuel and game_data.fuel.alive():
                target_type = "fuel"
            elif not player.in_car and (active_car or _alive_waiting_cars(game_data)):
                target_type = "car"
            else:
                target_type = None

            if target_type != hint_target_type:
                game_data.state.hint_target_type = target_type
                game_data.state.hint_expires_at = (
                    elapsed_ms + hint_delay if target_type else 0
                )
                hint_expires_at = game_data.state.hint_expires_at
                hint_target_type = target_type

            if (
                target_type
                and hint_expires_at
                and elapsed_ms >= hint_expires_at
                and not player.in_car
            ):
                if target_type == "fuel" and game_data.fuel and game_data.fuel.alive():
                    hint_target = game_data.fuel.rect.center
                elif target_type == "car":
                    if active_car:
                        hint_target = active_car.rect.center
                    else:
                        waiting_target = nearest_waiting_car(
                            game_data, (player.x, player.y)
                        )
                        if waiting_target:
                            hint_target = waiting_target.rect.center

        draw(
            render_assets,
            screen,
            game_data,
            config=config,
            hint_target=hint_target,
            hint_color=hint_color,
            fps=current_fps,
        )
        if profiling_active:
            font_settings = get_font_settings()
            font = load_font(font_settings.resource, font_settings.scaled_size(11))
            label = render_text_surface(
                font,
                "PROFILE ON",
                RED,
                line_height_scale=font_settings.line_height_scale,
            )
            screen.blit(label, (6, 6))
        present(screen)

    # Should not reach here, but return to title if it happens
    return _finalize(ScreenTransition(ScreenID.TITLE))



from __future__ import annotations

from typing import Tuple

import pygame

DEADZONE = 0.25
JOY_BUTTON_A = 0
JOY_BUTTON_BACK = 6
JOY_BUTTON_START = 7
JOY_BUTTON_RB = 5

CONTROLLER_AVAILABLE = hasattr(pygame, "controller")
CONTROLLER_DEVICE_ADDED = getattr(pygame, "CONTROLLERDEVICEADDED", None)
CONTROLLER_DEVICE_REMOVED = getattr(pygame, "CONTROLLERDEVICEREMOVED", None)
CONTROLLER_BUTTON_DOWN = getattr(pygame, "CONTROLLERBUTTONDOWN", None)
CONTROLLER_BUTTON_A = getattr(pygame, "CONTROLLER_BUTTON_A", None)
CONTROLLER_BUTTON_BACK = getattr(pygame, "CONTROLLER_BUTTON_BACK", None)
CONTROLLER_BUTTON_START = getattr(pygame, "CONTROLLER_BUTTON_START", None)
CONTROLLER_BUTTON_DPAD_UP = getattr(pygame, "CONTROLLER_BUTTON_DPAD_UP", None)
CONTROLLER_BUTTON_DPAD_DOWN = getattr(pygame, "CONTROLLER_BUTTON_DPAD_DOWN", None)
CONTROLLER_BUTTON_DPAD_LEFT = getattr(pygame, "CONTROLLER_BUTTON_DPAD_LEFT", None)
CONTROLLER_BUTTON_DPAD_RIGHT = getattr(pygame, "CONTROLLER_BUTTON_DPAD_RIGHT", None)
CONTROLLER_BUTTON_RB = getattr(pygame, "CONTROLLER_BUTTON_RIGHTSHOULDER", None)
CONTROLLER_AXIS_LEFTX = getattr(pygame, "CONTROLLER_AXIS_LEFTX", None)
CONTROLLER_AXIS_LEFTY = getattr(pygame, "CONTROLLER_AXIS_LEFTY", None)
CONTROLLER_AXIS_TRIGGERRIGHT = getattr(pygame, "CONTROLLER_AXIS_TRIGGERRIGHT", None)


def init_first_controller() -> pygame.controller.Controller | None:
    if not CONTROLLER_AVAILABLE:
        return None
    try:
        if pygame.controller.get_count() > 0:
            controller = pygame.controller.Controller(0)
            if not controller.get_init():
                controller.init()
            return controller
    except pygame.error:
        return None
    return None


def init_first_joystick() -> pygame.joystick.Joystick | None:
    try:
        if pygame.joystick.get_count() > 0:
            joystick = pygame.joystick.Joystick(0)
            if not joystick.get_init():
                joystick.init()
            return joystick
    except pygame.error:
        return None
    return None


def is_confirm_event(event: pygame.event.Event) -> bool:
    if CONTROLLER_BUTTON_DOWN is not None and event.type == CONTROLLER_BUTTON_DOWN:
        return CONTROLLER_BUTTON_A is not None and event.button == CONTROLLER_BUTTON_A
    if event.type == pygame.JOYBUTTONDOWN:
        return event.button == JOY_BUTTON_A
    return False


def is_confirm_held(
    controller: pygame.controller.Controller | None,
    joystick: pygame.joystick.Joystick | None,
) -> bool:
    """Return True if the confirm (South/A) button is currently held."""
    if controller and controller.get_init():
        if CONTROLLER_BUTTON_A is not None and controller.get_button(
            CONTROLLER_BUTTON_A
        ):
            return True
    if joystick and joystick.get_init():
        if joystick.get_numbuttons() > JOY_BUTTON_A and joystick.get_button(
            JOY_BUTTON_A
        ):
            return True
    return False


def is_start_event(event: pygame.event.Event) -> bool:
    if CONTROLLER_BUTTON_DOWN is not None and event.type == CONTROLLER_BUTTON_DOWN:
        return (
            CONTROLLER_BUTTON_START is not None
            and event.button == CONTROLLER_BUTTON_START
        )
    if event.type == pygame.JOYBUTTONDOWN:
        return event.button == JOY_BUTTON_START
    return False


def is_select_event(event: pygame.event.Event) -> bool:
    if CONTROLLER_BUTTON_DOWN is not None and event.type == CONTROLLER_BUTTON_DOWN:
        return (
            CONTROLLER_BUTTON_BACK is not None
            and event.button == CONTROLLER_BUTTON_BACK
        )
    if event.type == pygame.JOYBUTTONDOWN:
        return event.button == JOY_BUTTON_BACK
    return False


def read_gamepad_move(
    controller: pygame.controller.Controller | None,
    joystick: pygame.joystick.Joystick | None,
    *,
    deadzone: float = DEADZONE,
) -> Tuple[float, float]:
    x = 0.0
    y = 0.0
    if controller and controller.get_init():
        if CONTROLLER_AXIS_LEFTX is None or CONTROLLER_AXIS_LEFTY is None:
            return 0.0, 0.0
        x = float(controller.get_axis(CONTROLLER_AXIS_LEFTX))
        y = float(controller.get_axis(CONTROLLER_AXIS_LEFTY))
        if abs(x) < deadzone:
            x = 0.0
        if abs(y) < deadzone:
            y = 0.0
        if CONTROLLER_BUTTON_DPAD_LEFT is not None and controller.get_button(
            CONTROLLER_BUTTON_DPAD_LEFT
        ):
            x = -1.0
        elif CONTROLLER_BUTTON_DPAD_RIGHT is not None and controller.get_button(
            CONTROLLER_BUTTON_DPAD_RIGHT
        ):
            x = 1.0
        if CONTROLLER_BUTTON_DPAD_UP is not None and controller.get_button(
            CONTROLLER_BUTTON_DPAD_UP
        ):
            y = -1.0
        elif CONTROLLER_BUTTON_DPAD_DOWN is not None and controller.get_button(
            CONTROLLER_BUTTON_DPAD_DOWN
        ):
            y = 1.0
        return x, y

    if joystick and joystick.get_init():
        if joystick.get_numaxes() >= 2:
            x = float(joystick.get_axis(0))
            y = float(joystick.get_axis(1))
            if abs(x) < deadzone:
                x = 0.0
            if abs(y) < deadzone:
                y = 0.0
        if joystick.get_numhats() > 0:
            hat_x, hat_y = joystick.get_hat(0)
            if hat_x:
                x = float(hat_x)
            if hat_y:
                y = float(-hat_y)
    return x, y


def is_accel_active(
    keys: pygame.key.ScancodeWrapper,
    controller: pygame.controller.Controller | None,
    joystick: pygame.joystick.Joystick | None,
) -> bool:
    if keys[pygame.K_LSHIFT] or keys[pygame.K_RSHIFT]:
        return True
    if controller and controller.get_init():
        if CONTROLLER_BUTTON_RB is not None and controller.get_button(
            CONTROLLER_BUTTON_RB
        ):
            return True
        if CONTROLLER_AXIS_TRIGGERRIGHT is not None:
            if controller.get_axis(CONTROLLER_AXIS_TRIGGERRIGHT) > DEADZONE:
                return True
    if joystick and joystick.get_init():
        if joystick.get_numbuttons() > JOY_BUTTON_RB:
            if joystick.get_button(JOY_BUTTON_RB):
                return True
        if joystick.get_numaxes() > 5:
            if joystick.get_axis(5) > DEADZONE:
                return True
    return False



from __future__ import annotations

import pygame
from pygame import sprite, surface

from ..colors import (
    BLACK,
    BLUE,
    FOOTPRINT_COLOR,
    YELLOW,
    WHITE,
    get_environment_palette,
)
from ..entities import (
    Car,
    Flashlight,
    FuelCan,
    Player,
    PatrolBot,
    Shoes,
    SteelBeam,
    Survivor,
    Wall,
    Zombie,
    ZombieDog,
)
from ..font_utils import load_font, render_text_surface
from ..localization import get_font_settings
from ..models import Footprint, GameData
from ..render_assets import RenderAssets, resolve_steel_beam_colors, resolve_wall_colors
from ..render_constants import MOVING_FLOOR_OVERVIEW_COLOR
from ..entities_constants import PATROL_BOT_COLLISION_RADIUS
from .hud import _get_fog_scale


def compute_floor_cells(
    *,
    cols: int,
    rows: int,
    wall_cells: set[tuple[int, int]],
    outer_wall_cells: set[tuple[int, int]],
    pitfall_cells: set[tuple[int, int]],
) -> set[tuple[int, int]]:
    """Return floor cells for the minimap base pass."""
    # The layout wall sets are updated when walls are destroyed, so removing
    # those cells here makes the minimap treat destroyed walls as floor.
    blocked = wall_cells | outer_wall_cells | pitfall_cells
    return {(x, y) for y in range(rows) for x in range(cols) if (x, y) not in blocked}


def _draw_overview_tag(
    surface: surface.Surface,
    font: pygame.font.Font,
    text: str,
    item_rect: pygame.Rect,
    *,
    fg: tuple[int, int, int] = YELLOW,
    padding: tuple[int, int] = (4, 2),
    line_height_scale: float = 1.0,
) -> None:
    label = render_text_surface(
        font, text, fg, line_height_scale=line_height_scale
    )
    label_rect = label.get_rect()
    padded = label_rect.inflate(padding[0] * 2, padding[1] * 2)
    top_left = (item_rect.left, item_rect.top)
    bottom_left = (item_rect.left, item_rect.bottom - padded.height)
    if top_left[1] < 0 or top_left[1] + padded.height > surface.get_height():
        x, y = bottom_left
    else:
        x, y = top_left
    x = max(0, min(surface.get_width() - padded.width, x))
    y = max(0, min(surface.get_height() - padded.height, y))
    padded.topleft = (x, y)
    label_rect.center = padded.center
    surface.blit(label, label_rect)


def draw_level_overview(
    assets: RenderAssets,
    surface: surface.Surface,
    wall_group: sprite.Group,
    floor_cells: set[tuple[int, int]],
    player: Player | None,
    car: Car | None,
    waiting_cars: list[Car] | None,
    footprints: list[Footprint],
    *,
    now_ms: int,
    fuel: FuelCan | None = None,
    flashlights: list[Flashlight] | None = None,
    shoes: list[Shoes] | None = None,
    buddies: list[Survivor] | None = None,
    survivors: list[Survivor] | None = None,
    patrol_bots: list[PatrolBot] | None = None,
    zombies: list[pygame.sprite.Sprite] | None = None,
    fall_spawn_cells: set[tuple[int, int]] | None = None,
    moving_floor_cells: dict[tuple[int, int], object] | None = None,
    palette_key: str | None = None,
) -> None:
    palette = get_environment_palette(palette_key)
    base_floor = palette.floor_primary
    dark_floor = tuple(max(0, int(channel * 0.35)) for channel in base_floor)
    floor_color = tuple(max(0, int(channel * 0.65)) for channel in base_floor)
    fall_floor = tuple(
        max(0, int(channel * 0.55)) for channel in palette.fall_zone_primary
    )
    surface.fill(dark_floor)
    cell_size = assets.internal_wall_grid_snap
    if cell_size > 0:
        for x, y in floor_cells:
            pygame.draw.rect(
                surface,
                floor_color,
                pygame.Rect(
                    x * cell_size,
                    y * cell_size,
                    cell_size,
                    cell_size,
                ),
            )
        if fall_spawn_cells:
            for x, y in fall_spawn_cells:
                pygame.draw.rect(
                    surface,
                    fall_floor,
                    pygame.Rect(
                        x * cell_size,
                        y * cell_size,
                        cell_size,
                        cell_size,
                    ),
                )
        if moving_floor_cells:
            for x, y in moving_floor_cells.keys():
                pygame.draw.rect(
                    surface,
                    MOVING_FLOOR_OVERVIEW_COLOR,
                    pygame.Rect(
                        x * cell_size,
                        y * cell_size,
                        cell_size,
                        cell_size,
                    ),
                )

    for wall in wall_group:
        if wall.max_health > 0:
            health_ratio = max(0.0, min(1.0, wall.health / wall.max_health))
        else:
            health_ratio = 0.0
        if isinstance(wall, Wall):
            if health_ratio <= 0.0:
                pygame.draw.rect(surface, floor_color, wall.rect)
            else:
                fill_color, _ = resolve_wall_colors(
                    health_ratio=health_ratio,
                    palette_category=wall.palette_category,
                    palette=palette,
                )
                pygame.draw.rect(surface, fill_color, wall.rect)
        elif isinstance(wall, SteelBeam):
            if health_ratio <= 0.0:
                pygame.draw.rect(surface, floor_color, wall.rect)
            else:
                fill_color, _ = resolve_steel_beam_colors(
                    health_ratio=health_ratio,
                    palette=palette,
                )
                pygame.draw.rect(surface, fill_color, wall.rect)
    now = now_ms
    for fp in footprints:
        if not fp.visible:
            continue
        age = now - fp.time
        fade = 1 - (age / assets.footprint_lifetime_ms)
        fade = max(assets.footprint_min_fade, fade)
        color = tuple(max(0, min(255, int(c * fade))) for c in FOOTPRINT_COLOR)
        pygame.draw.circle(
            surface,
            color,
            (int(fp.pos[0]), int(fp.pos[1])),
            assets.footprint_overview_radius,
        )
    if fuel and fuel.alive():
        pygame.draw.rect(surface, YELLOW, fuel.rect, border_radius=3)
        pygame.draw.rect(surface, BLACK, fuel.rect, width=2, border_radius=3)
    if flashlights:
        for flashlight in flashlights:
            if flashlight.alive():
                pygame.draw.rect(surface, YELLOW, flashlight.rect, border_radius=2)
                pygame.draw.rect(
                    surface, BLACK, flashlight.rect, width=2, border_radius=2
                )
    if shoes:
        for item in shoes:
            if item.alive():
                surface.blit(item.image, item.rect)
    if survivors:
        for survivor in survivors:
            if survivor.alive():
                pygame.draw.circle(
                    surface,
                    (220, 220, 255),
                    survivor.rect.center,
                    assets.player_radius * 2,
                )
    if player:
        pygame.draw.circle(surface, BLUE, player.rect.center, assets.player_radius * 2)
    if buddies:
        buddy_color = (0, 200, 70)
        for buddy in buddies:
            if buddy.alive() and not buddy.rescued:
                pygame.draw.circle(
                    surface, buddy_color, buddy.rect.center, assets.player_radius * 2
                )
    drawn_cars: list[Car] = []
    if car and car.alive():
        car_rect = car.image.get_rect(center=car.rect.center)
        surface.blit(car.image, car_rect)
        drawn_cars.append(car)
    if waiting_cars:
        for parked in waiting_cars:
            if not parked.alive() or parked in drawn_cars:
                continue
            parked_rect = parked.image.get_rect(center=parked.rect.center)
            surface.blit(parked.image, parked_rect)
    if patrol_bots:
        for bot in patrol_bots:
            if bot.alive():
                pygame.draw.circle(
                    surface,
                    (90, 45, 120),
                    bot.rect.center,
                    int(PATROL_BOT_COLLISION_RADIUS),
                )
    if zombies:
        zombie_color = (200, 80, 80)
        zombie_radius = max(2, int(assets.player_radius * 1.2))
        for zombie in zombies:
            if not zombie.alive():
                continue
            if isinstance(zombie, (Zombie, ZombieDog)):
                pygame.draw.circle(
                    surface,
                    zombie_color,
                    zombie.rect.center,
                    zombie_radius,
                )


def draw_debug_overview(
    assets: RenderAssets,
    screen: surface.Surface,
    overview_surface: surface.Surface,
    game_data: GameData,
    config: dict[str, object],
    *,
    screen_width: int,
    screen_height: int,
) -> None:
    cell_size = assets.internal_wall_grid_snap
    floor_cells: set[tuple[int, int]] = set()
    if cell_size > 0:
        floor_cells = compute_floor_cells(
            cols=max(0, game_data.layout.field_rect.width // cell_size),
            rows=max(0, game_data.layout.field_rect.height // cell_size),
            wall_cells=game_data.layout.wall_cells,
            outer_wall_cells=game_data.layout.outer_wall_cells,
            pitfall_cells=game_data.layout.pitfall_cells,
        )
    footprints_enabled = bool(config.get("footprints", {}).get("enabled", True))
    footprints_to_draw = game_data.state.footprints if footprints_enabled else []
    draw_level_overview(
        assets,
        overview_surface,
        game_data.groups.wall_group,
        floor_cells,
        game_data.player,
        game_data.car,
        game_data.waiting_cars,
        footprints_to_draw,
        now_ms=game_data.state.clock.elapsed_ms,
        fuel=game_data.fuel,
        flashlights=game_data.flashlights or [],
        shoes=game_data.shoes or [],
        buddies=[
            survivor
            for survivor in game_data.groups.survivor_group
            if survivor.alive() and survivor.is_buddy and not survivor.rescued
        ],
        survivors=list(game_data.groups.survivor_group),
        patrol_bots=list(game_data.groups.patrol_bot_group),
        zombies=list(game_data.groups.zombie_group),
        fall_spawn_cells=game_data.layout.fall_spawn_cells,
        moving_floor_cells=game_data.layout.moving_floor_cells,
        palette_key=game_data.state.ambient_palette_key,
    )
    fov_target = None
    if (
        game_data.player
        and game_data.player.in_car
        and game_data.car
        and game_data.car.alive()
    ):
        fov_target = game_data.car
    elif game_data.player:
        fov_target = game_data.player
    if fov_target:
        fov_scale = _get_fog_scale(assets, game_data.state.flashlight_count)
        fov_radius = max(1, int(assets.fov_radius * fov_scale))
        pygame.draw.circle(
            overview_surface,
            (255, 255, 120),
            fov_target.rect.center,
            fov_radius,
            width=2,
        )
    cam_offset = game_data.camera.camera
    camera_rect = pygame.Rect(
        -cam_offset.x,
        -cam_offset.y,
        screen_width,
        screen_height,
    )
    pygame.draw.rect(overview_surface, WHITE, camera_rect, width=1)
    level_rect = game_data.layout.field_rect
    level_aspect = level_rect.width / max(1, level_rect.height)
    screen_aspect = screen_width / max(1, screen_height)
    if level_aspect > screen_aspect:
        scaled_w = screen_width - 40
        scaled_h = int(scaled_w / level_aspect)
    else:
        scaled_h = screen_height - 40
        scaled_w = int(scaled_h * level_aspect)
    scaled_w = max(1, scaled_w)
    scaled_h = max(1, scaled_h)
    scaled_overview = pygame.transform.smoothscale(
        overview_surface, (scaled_w, scaled_h)
    )
    screen.fill(BLACK)
    scaled_rect = scaled_overview.get_rect(
        center=(screen_width // 2, screen_height // 2)
    )
    screen.blit(
        scaled_overview,
        scaled_rect,
    )
    try:
        font_settings = get_font_settings()
        label_font = load_font(font_settings.resource, font_settings.scaled_size(11))
    except pygame.error as e:
        print(f"Error loading overview font: {e}")
        return
    scale_x = scaled_w / max(1, level_rect.width)
    scale_y = scaled_h / max(1, level_rect.height)

    def _scaled_rect(rect: pygame.Rect) -> pygame.Rect:
        return pygame.Rect(
            int(scaled_rect.left + rect.left * scale_x),
            int(scaled_rect.top + rect.top * scale_y),
            max(1, int(rect.width * scale_x)),
            max(1, int(rect.height * scale_y)),
        )

    if game_data.car and game_data.car.alive():
        _draw_overview_tag(
            screen,
            label_font,
            "C",
            _scaled_rect(game_data.car.rect),
            line_height_scale=font_settings.line_height_scale,
        )
    for parked in game_data.waiting_cars:
        if parked.alive():
            _draw_overview_tag(
                screen,
                label_font,
                "C",
                _scaled_rect(parked.rect),
                line_height_scale=font_settings.line_height_scale,
            )
    if game_data.fuel and game_data.fuel.alive():
        _draw_overview_tag(
            screen,
            label_font,
            "F",
            _scaled_rect(game_data.fuel.rect),
            line_height_scale=font_settings.line_height_scale,
        )
    if game_data.flashlights:
        for flashlight in game_data.flashlights:
            if flashlight.alive():
                _draw_overview_tag(
                    screen,
                    label_font,
                    "L",
                    _scaled_rect(flashlight.rect),
                    line_height_scale=font_settings.line_height_scale,
                )
    if game_data.shoes:
        for item in game_data.shoes:
            if item.alive():
                _draw_overview_tag(
                    screen,
                    label_font,
                    "S",
                    _scaled_rect(item.rect),
                    line_height_scale=font_settings.line_height_scale,
                )



from __future__ import annotations

import webbrowser
from typing import Any, Sequence

import pygame
from pygame import surface, time

from ..colors import BLACK, GRAY, LIGHT_GRAY, WHITE
from ..font_utils import load_font
from ..localization import get_font_settings, get_language
from ..localization import translate as tr
from ..models import Stage
from ..progress import load_progress
from ..render import blit_text_wrapped, wrap_text
from ..rng import generate_seed
from ..input_utils import (
    CONTROLLER_BUTTON_DOWN,
    CONTROLLER_BUTTON_DPAD_DOWN,
    CONTROLLER_BUTTON_DPAD_LEFT,
    CONTROLLER_BUTTON_DPAD_RIGHT,
    CONTROLLER_BUTTON_DPAD_UP,
    CONTROLLER_DEVICE_ADDED,
    CONTROLLER_DEVICE_REMOVED,
    init_first_controller,
    init_first_joystick,
    is_confirm_event,
)
from ..screens import (
    ScreenID,
    ScreenTransition,
    TITLE_HEADER_Y,
    TITLE_SECTION_TOP,
)
from ..windowing import (
    adjust_menu_logical_size,
    nudge_menu_window_scale,
    present,
    sync_window_size,
    toggle_fullscreen,
)

try:  # pragma: no cover - version fallback not critical for tests
    from ..__about__ import __version__
except Exception:  # pragma: no cover - fallback version
    __version__ = "0.0.0-unknown"

MAX_SEED_DIGITS = 19
_README_URLS: dict[str, str] = {
    "en": "https://github.com/tos-kamiya/zombie-escape/blob/main/README.md",
    "ja": "https://github.com/tos-kamiya/zombie-escape/blob/main/README-ja_JP.md",
}
_STAGE6_URLS: dict[str, str] = {
    "en": "https://github.com/tos-kamiya/zombie-escape/blob/main/docs/stages-6plus.md",
    "ja": "https://github.com/tos-kamiya/zombie-escape/blob/main/docs/stages-6plus-ja_JP.md",
}
_UNCLEARED_STAGE_COLOR: tuple[int, int, int] = (220, 80, 80)


def _open_readme_link(*, use_stage6: bool = False) -> None:
    """Open the GitHub README or Stage 6+ guide for the active UI language."""

    language = get_language()
    if use_stage6:
        url = _STAGE6_URLS.get(language, _STAGE6_URLS["en"])
    else:
        url = _README_URLS.get(language, _README_URLS["en"])
    try:
        webbrowser.open(url, new=0, autoraise=True)
    except Exception as exc:  # pragma: no cover - best effort only
        print(f"Unable to open README URL {url}: {exc}")


def _generate_auto_seed_text() -> str:
    raw = generate_seed()
    trimmed = raw // 100  # drop lower 2 digits for stability
    return str(trimmed % 100000).zfill(5)


def title_screen(
    screen: surface.Surface,
    clock: time.Clock,
    config: dict[str, Any],
    fps: int,
    *,
    stages: Sequence[Stage],
    default_stage_id: str,
    screen_size: tuple[int, int],
    seed_text: str | None = None,
    seed_is_auto: bool = False,
) -> ScreenTransition:
    """Display the title menu and return the selected transition."""

    width, height = screen.get_size()
    if width <= 0 or height <= 0:
        width, height = screen_size
    stage_options_all: list[dict] = [
        {"type": "stage", "stage": stage, "available": stage.available}
        for stage in stages
        if stage.available
    ]
    first_page_size = 5
    other_page_size = 10
    stage_pages: list[list[dict]] = []
    if stage_options_all:
        stage_pages.append(stage_options_all[:first_page_size])
        for i in range(first_page_size, len(stage_options_all), other_page_size):
            stage_pages.append(stage_options_all[i : i + other_page_size])
    resource_options: list[dict[str, Any]] = [
        {"type": "settings"},
        {"type": "readme"},
        {"type": "quit"},
    ]
    generated = seed_text is None
    current_seed_text = (
        seed_text if seed_text is not None else _generate_auto_seed_text()
    )
    current_seed_auto = seed_is_auto or generated
    stage_progress, _ = load_progress()

    def _page_available(page_index: int) -> bool:
        if page_index <= 0:
            return True
        required = stage_options_all[:first_page_size]
        return all(stage_progress.get(option["stage"].id, 0) > 0 for option in required)

    def _page_index_for_stage(stage_idx: int) -> int:
        if stage_idx < first_page_size:
            return 0
        return 1 + (stage_idx - first_page_size) // other_page_size

    current_page = 0
    if stage_options_all:
        for idx, opt in enumerate(stage_options_all):
            if opt["stage"].id == default_stage_id:
                target_page = _page_index_for_stage(idx)
                if _page_available(target_page):
                    current_page = target_page
                break

    def _build_options(page_index: int) -> tuple[list[dict], list[dict]]:
        page_index = max(0, min(page_index, len(stage_pages) - 1))
        stage_options = stage_pages[page_index] if stage_pages else []
        options = list(stage_options) + resource_options
        return options, stage_options

    options, stage_options = _build_options(current_page)
    selected_stage_index = next(
        (
            i
            for i, opt in enumerate(options)
            if opt["type"] == "stage" and opt["stage"].id == default_stage_id
        ),
        0,
    )
    selected = min(selected_stage_index, len(options) - 1)
    controller = init_first_controller()
    joystick = init_first_joystick() if controller is None else None
    pygame.event.pump()
    # Drop any queued confirm events from startup/controller init glitches.
    confirm_event_types = [pygame.JOYBUTTONDOWN]
    if CONTROLLER_BUTTON_DOWN is not None:
        confirm_event_types.append(CONTROLLER_BUTTON_DOWN)
    pygame.event.clear(confirm_event_types)
    pygame.event.clear([pygame.KEYDOWN])
    confirm_armed_at = pygame.time.get_ticks() + 300

    def _render_frame() -> None:
        screen.fill(BLACK)
        try:
            font_settings = get_font_settings()

            def _get_font(size: int) -> pygame.font.Font:
                return load_font(font_settings.resource, size)

            def _measure_text(
                text: str, font: pygame.font.Font, max_width: int
            ) -> tuple[int, int, int]:
                lines = wrap_text(text, font, max_width)
                line_height = int(
                    round(font.get_linesize() * font_settings.line_height_scale)
                )
                height = max(1, len(lines)) * line_height
                width = max((font.size(line)[0] for line in lines if line), default=0)
                return width, height, line_height

            base_stage_size = font_settings.scaled_size(11)
            selected_stage_size = font_settings.scaled_size(22)
            line_height_scale = font_settings.line_height_scale
            base_row_height = (
                int(
                    round(_get_font(base_stage_size).get_linesize() * line_height_scale)
                )
                + 2
            )
            selected_row_height = (
                int(
                    round(
                        _get_font(selected_stage_size).get_linesize()
                        * line_height_scale
                    )
                )
                + 2
            )
            list_column_x = 24
            list_column_width = width // 2 - 36
            info_column_x = width // 2 + 12
            info_column_width = width - info_column_x - 24
            section_top = TITLE_SECTION_TOP
            highlight_color = (70, 70, 70)

            stage_count = len(stage_options)
            # resource_count = len(options) - stage_count

            stage_header_text = tr("menu.sections.stage_select")
            show_page_arrows = len(stage_pages) > 1 and _page_available(1)
            if show_page_arrows:
                left_arrow = "<- " if current_page > 0 else ""
                right_arrow = (
                    " ->"
                    if current_page < len(stage_pages) - 1
                    and _page_available(current_page + 1)
                    else ""
                )
                stage_header_text = f"{left_arrow}{stage_header_text}{right_arrow}"
            section_size = font_settings.scaled_size(11)
            section_font = _get_font(section_size)
            header_width, header_height, _ = _measure_text(
                stage_header_text, section_font, list_column_width
            )
            blit_text_wrapped(
                screen,
                stage_header_text,
                section_font,
                LIGHT_GRAY,
                (list_column_x, section_top),
                list_column_width,
                line_height_scale=font_settings.line_height_scale,
            )
            stage_header_rect = pygame.Rect(
                list_column_x, section_top, max(header_width, 1), header_height
            )
            stage_rows_start = stage_header_rect.bottom + 6
            resource_row_height = base_row_height
            resource_offset = resource_row_height
            stage_row_heights = [
                (selected_row_height if idx == selected else base_row_height)
                for idx in range(stage_count)
            ]
            fixed_stage_block_height = (
                base_row_height * max(stage_count - 1, 0) + selected_row_height
            )
            action_header_pos = (
                list_column_x,
                stage_rows_start + fixed_stage_block_height + 14 + resource_offset,
            )
            action_header_text = tr("menu.sections.resources")
            action_width, action_height, _ = _measure_text(
                action_header_text, section_font, list_column_width
            )
            blit_text_wrapped(
                screen,
                action_header_text,
                section_font,
                LIGHT_GRAY,
                action_header_pos,
                list_column_width,
                line_height_scale=font_settings.line_height_scale,
            )
            action_header_rect = pygame.Rect(
                action_header_pos[0], action_header_pos[1], max(action_width, 1), action_height
            )
            action_rows_start = action_header_rect.bottom + 6

            row_top = stage_rows_start
            for idx, option in enumerate(stage_options):
                row_height = stage_row_heights[idx]
                highlight_rect = pygame.Rect(
                    list_column_x, row_top - 2, list_column_width, row_height
                )
                cleared = stage_progress.get(option["stage"].id, 0) > 0
                base_color = WHITE if cleared else _UNCLEARED_STAGE_COLOR
                color = base_color
                is_selected = idx == selected
                if is_selected:
                    pygame.draw.rect(screen, highlight_color, highlight_rect)
                label = option["stage"].name
                if not option.get("available"):
                    locked_suffix = tr("menu.locked_suffix")
                    label = f"{label} {locked_suffix}"
                    color = GRAY
                stage_option_size = (
                    selected_stage_size if is_selected else base_stage_size
                )
                stage_option_font = _get_font(stage_option_size)
                text_height = int(
                    round(
                        stage_option_font.get_linesize()
                        * font_settings.line_height_scale
                    )
                )
                blit_text_wrapped(
                    screen,
                    label,
                    stage_option_font,
                    color,
                    (
                        list_column_x + 8,
                        row_top + (row_height - text_height) // 2,
                    ),
                    10_000,
                    line_height_scale=font_settings.line_height_scale,
                )
                row_top += row_height

            resource_option_size = font_settings.scaled_size(11)
            resource_option_font = _get_font(resource_option_size)
            for idx, option in enumerate(resource_options):
                option_idx = stage_count + idx
                row_top = action_rows_start + idx * resource_row_height
                highlight_rect = pygame.Rect(
                    list_column_x, row_top - 2, list_column_width, resource_row_height
                )
                is_selected = option_idx == selected
                if is_selected:
                    pygame.draw.rect(screen, highlight_color, highlight_rect)
                if option["type"] == "settings":
                    label = tr("menu.settings")
                elif option["type"] == "readme":
                    label_key = (
                        "menu.readme_stage6" if current_page > 0 else "menu.readme"
                    )
                    label = f"> {tr(label_key)}"
                else:
                    label = tr("menu.quit")
                color = WHITE
                text_height = int(
                    round(
                        resource_option_font.get_linesize()
                        * font_settings.line_height_scale
                    )
                )
                blit_text_wrapped(
                    screen,
                    label,
                    resource_option_font,
                    color,
                    (
                        list_column_x + 8,
                        row_top + (resource_row_height - text_height) // 2,
                    ),
                    list_column_width - 12,
                    line_height_scale=font_settings.line_height_scale,
                )

            current = options[selected]
            desc_area_top = section_top
            if current["type"] == "stage":
                desc_size = font_settings.scaled_size(11)
                desc_font = _get_font(desc_size)
                desc_color = WHITE if current.get("available") else GRAY
                desc_lines = wrap_text(
                    current["stage"].description, desc_font, info_column_width
                )
                _, _, desc_line_height = _measure_text(
                    current["stage"].description, desc_font, info_column_width
                )
                desc_height = max(1, len(desc_lines)) * desc_line_height
                desc_panel_padding = 6
                desc_panel_rect = pygame.Rect(
                    info_column_x - desc_panel_padding,
                    desc_area_top - desc_panel_padding,
                    info_column_width + desc_panel_padding * 2,
                    desc_height + desc_panel_padding * 2,
                )
                desc_panel = pygame.Surface(
                    (desc_panel_rect.width, desc_panel_rect.height), pygame.SRCALPHA
                )
                desc_panel.fill((0, 0, 0, 140))
                screen.blit(desc_panel, desc_panel_rect.topleft)
                blit_text_wrapped(
                    screen,
                    current["stage"].description,
                    desc_font,
                    desc_color,
                    (info_column_x, desc_area_top),
                    info_column_width,
                    line_height_scale=font_settings.line_height_scale,
                )

            option_help_top = desc_area_top
            help_text = ""
            if current["type"] == "settings":
                help_text = tr("menu.option_help.settings")
            elif current["type"] == "quit":
                help_text = tr("menu.option_help.quit")
            elif current["type"] == "readme":
                help_key = (
                    "menu.option_help.readme_stage6"
                    if current_page > 0
                    else "menu.option_help.readme"
                )
                help_text = tr(help_key)

            if help_text:
                desc_size = font_settings.scaled_size(11)
                desc_font = _get_font(desc_size)
                blit_text_wrapped(
                    screen,
                    help_text,
                    desc_font,
                    WHITE,
                    (info_column_x, option_help_top),
                    info_column_width,
                    line_height_scale=font_settings.line_height_scale,
                )

            hint_size = font_settings.scaled_size(11)
            hint_font = _get_font(hint_size)
            hint_line_height = int(
                round(hint_font.get_linesize() * font_settings.line_height_scale)
            )
            hint_start_y = action_header_pos[1]
            hint_step = hint_line_height
            if current_page == 0:
                hint_lines = [tr("menu.hints.navigate")]
                if len(stage_pages) > 1 and _page_available(1):
                    hint_lines.append(tr("menu.hints.page_switch"))
                hint_lines.extend(tr("menu.hints.confirm").splitlines())
                for offset, line in enumerate(hint_lines):
                    blit_text_wrapped(
                        screen,
                        line,
                        hint_font,
                        WHITE,
                        (info_column_x, hint_start_y + offset * hint_step),
                        info_column_width,
                        line_height_scale=font_settings.line_height_scale,
                    )

            seed_value_display = (
                current_seed_text if current_seed_text else tr("menu.seed_empty")
            )
            seed_label = tr("menu.seed_label", value=seed_value_display)
            seed_offset_y = hint_step
            seed_width, seed_height, _ = _measure_text(
                seed_label, hint_font, info_column_width
            )
            seed_bottom = height - 30 + seed_offset_y
            blit_text_wrapped(
                screen,
                seed_label,
                hint_font,
                LIGHT_GRAY,
                (info_column_x, seed_bottom - seed_height),
                info_column_width,
                line_height_scale=font_settings.line_height_scale,
            )
            seed_rect = pygame.Rect(
                info_column_x, seed_bottom - seed_height, max(seed_width, 1), seed_height
            )

            if current_page == 0:
                seed_hint = tr("menu.seed_hint")
                seed_hint_lines = wrap_text(seed_hint, hint_font, info_column_width)
                seed_hint_height = len(seed_hint_lines) * hint_line_height
                seed_hint_top = seed_rect.top - 4 - seed_hint_height
                blit_text_wrapped(
                    screen,
                    seed_hint,
                    hint_font,
                    LIGHT_GRAY,
                    (info_column_x, seed_hint_top),
                    info_column_width,
                    line_height_scale=font_settings.line_height_scale,
                )

            title_text = tr("game.title")
            title_font = _get_font(font_settings.scaled_size(33))
            title_width, title_height, _ = _measure_text(
                title_text, title_font, width
            )
            title_topleft = (
                width // 2 - title_width // 2,
                TITLE_HEADER_Y - title_height // 2,
            )
            blit_text_wrapped(
                screen,
                title_text,
                title_font,
                LIGHT_GRAY,
                title_topleft,
                width,
                line_height_scale=font_settings.line_height_scale,
            )
            title_rect = pygame.Rect(
                title_topleft[0], title_topleft[1], max(title_width, 1), title_height
            )
            version_font = _get_font(font_settings.scaled_size(11))
            version_text = f"v{__version__}"
            version_width, version_height, _ = _measure_text(
                version_text, version_font, width
            )
            version_topleft = (
                title_rect.right + 4,
                title_rect.bottom - version_height,
            )
            blit_text_wrapped(
                screen,
                version_text,
                version_font,
                LIGHT_GRAY,
                version_topleft,
                width - version_topleft[0],
                line_height_scale=font_settings.line_height_scale,
            )

        except pygame.error as e:
            print(f"Error rendering title screen: {e}")

    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return ScreenTransition(
                    ScreenID.EXIT,
                    seed_text=current_seed_text,
                    seed_is_auto=current_seed_auto,
                )
            if event.type in (pygame.WINDOWSIZECHANGED, pygame.VIDEORESIZE):
                sync_window_size(event)
                adjust_menu_logical_size()
                continue
            if event.type == pygame.JOYDEVICEADDED or (
                CONTROLLER_DEVICE_ADDED is not None
                and event.type == CONTROLLER_DEVICE_ADDED
            ):
                if controller is None:
                    controller = init_first_controller()
                if controller is None:
                    joystick = init_first_joystick()
            if event.type == pygame.JOYDEVICEREMOVED or (
                CONTROLLER_DEVICE_REMOVED is not None
                and event.type == CONTROLLER_DEVICE_REMOVED
            ):
                if controller and not controller.get_init():
                    controller = None
                if joystick and not joystick.get_init():
                    joystick = None
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_BACKSPACE:
                    current_seed_text = _generate_auto_seed_text()
                    current_seed_auto = True
                    continue
                if event.unicode and event.unicode.isdigit():
                    if current_seed_auto:
                        current_seed_text = ""
                        current_seed_auto = False
                    if len(current_seed_text) < MAX_SEED_DIGITS:
                        current_seed_text += event.unicode
                    continue
                if event.key == pygame.K_LEFTBRACKET:
                    nudge_menu_window_scale(0.5)
                    continue
                if event.key == pygame.K_RIGHTBRACKET:
                    nudge_menu_window_scale(2.0)
                    continue
                if event.key == pygame.K_f:
                    toggle_fullscreen()
                    adjust_menu_logical_size()
                    continue
                if event.key == pygame.K_LEFT:
                    if current_page > 0:
                        current_page -= 1
                        options, stage_options = _build_options(current_page)
                        selected = 0
                    continue
                if event.key == pygame.K_RIGHT:
                    if current_page < len(stage_pages) - 1 and _page_available(
                        current_page + 1
                    ):
                        current_page += 1
                        options, stage_options = _build_options(current_page)
                        selected = 0
                    continue
                if event.key in (pygame.K_UP, pygame.K_w):
                    selected = (selected - 1) % len(options)
                elif event.key in (pygame.K_DOWN, pygame.K_s):
                    selected = (selected + 1) % len(options)
                elif event.key in (pygame.K_RETURN, pygame.K_SPACE):
                    if pygame.time.get_ticks() < confirm_armed_at:
                        continue
                    current = options[selected]
                    if current["type"] == "stage" and current.get("available"):
                        seed_value = (
                            int(current_seed_text) if current_seed_text else None
                        )
                        return ScreenTransition(
                            ScreenID.GAMEPLAY,
                            stage=current["stage"],
                            seed=seed_value,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                    if current["type"] == "settings":
                        return ScreenTransition(
                            ScreenID.SETTINGS,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                    if current["type"] == "readme":
                        _open_readme_link(use_stage6=current_page > 0)
                        continue
                    if current["type"] == "quit":
                        return ScreenTransition(
                            ScreenID.EXIT,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
            if event.type == pygame.JOYBUTTONDOWN or (
                CONTROLLER_BUTTON_DOWN is not None
                and event.type == CONTROLLER_BUTTON_DOWN
            ):
                if is_confirm_event(event):
                    if pygame.time.get_ticks() < confirm_armed_at:
                        continue
                    current = options[selected]
                    if current["type"] == "stage" and current.get("available"):
                        seed_value = (
                            int(current_seed_text) if current_seed_text else None
                        )
                        return ScreenTransition(
                            ScreenID.GAMEPLAY,
                            stage=current["stage"],
                            seed=seed_value,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                    if current["type"] == "settings":
                        return ScreenTransition(
                            ScreenID.SETTINGS,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                    if current["type"] == "readme":
                        _open_readme_link(use_stage6=current_page > 0)
                        continue
                    if current["type"] == "quit":
                        return ScreenTransition(
                            ScreenID.EXIT,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                if (
                    CONTROLLER_BUTTON_DOWN is not None
                    and event.type == CONTROLLER_BUTTON_DOWN
                ):
                    if (
                        CONTROLLER_BUTTON_DPAD_UP is not None
                        and event.button == CONTROLLER_BUTTON_DPAD_UP
                    ):
                        selected = (selected - 1) % len(options)
                    if (
                        CONTROLLER_BUTTON_DPAD_DOWN is not None
                        and event.button == CONTROLLER_BUTTON_DPAD_DOWN
                    ):
                        selected = (selected + 1) % len(options)
                    if (
                        CONTROLLER_BUTTON_DPAD_LEFT is not None
                        and event.button == CONTROLLER_BUTTON_DPAD_LEFT
                    ):
                        if current_page > 0:
                            current_page -= 1
                            options, stage_options = _build_options(current_page)
                            selected = 0
                    if (
                        CONTROLLER_BUTTON_DPAD_RIGHT is not None
                        and event.button == CONTROLLER_BUTTON_DPAD_RIGHT
                    ):
                        if current_page < len(stage_pages) - 1 and _page_available(
                            current_page + 1
                        ):
                            current_page += 1
                            options, stage_options = _build_options(current_page)
                            selected = 0
            if event.type == pygame.JOYHATMOTION:
                hat_x, hat_y = event.value
                if hat_y == 1:
                    selected = (selected - 1) % len(options)
                elif hat_y == -1:
                    selected = (selected + 1) % len(options)
                if hat_x == -1:
                    if current_page > 0:
                        current_page -= 1
                        options, stage_options = _build_options(current_page)
                        selected = 0
                elif hat_x == 1:
                    if current_page < len(stage_pages) - 1 and _page_available(
                        current_page + 1
                    ):
                        current_page += 1
                        options, stage_options = _build_options(current_page)
                        selected = 0

        _render_frame()
        present(screen)
        clock.tick(fps)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable

import pygame
import numpy as np  # type: ignore

from ..entities_constants import (
    ZOMBIE_DOG_LONG_AXIS_RATIO,
    ZOMBIE_DOG_SHORT_AXIS_RATIO,
    ZOMBIE_RADIUS,
)
from ..render_assets import (
    build_zombie_directional_surfaces,
    build_zombie_dog_directional_surfaces,
)
from ..screen_constants import FPS
from ..rng import get_rng

RNG = get_rng()


DECAY_EFFECT_DURATION_FRAMES = max(1, int(FPS))


@dataclass(frozen=True)
class DecayMask:
    width: int
    height: int
    positions: list[tuple[int, int]]


_DECAY_MASK: DecayMask | None = None


def _max_decay_surface_size() -> tuple[int, int]:
    zombie_surfaces = build_zombie_directional_surfaces(ZOMBIE_RADIUS, draw_hands=False)
    base_size = ZOMBIE_RADIUS * 2.0
    long_axis = base_size * ZOMBIE_DOG_LONG_AXIS_RATIO
    short_axis = base_size * ZOMBIE_DOG_SHORT_AXIS_RATIO
    dog_surfaces = build_zombie_dog_directional_surfaces(long_axis, short_axis)
    widths = [surf.get_width() for surf in zombie_surfaces + dog_surfaces]
    heights = [surf.get_height() for surf in zombie_surfaces + dog_surfaces]
    return max(widths), max(heights)


def _build_decay_mask() -> DecayMask:
    width, height = _max_decay_surface_size()
    positions = [(x, y) for y in range(height) for x in range(width)]
    RNG.shuffle(positions)
    return DecayMask(width=width, height=height, positions=positions)


def prepare_decay_mask() -> DecayMask:
    global _DECAY_MASK
    if _DECAY_MASK is None:
        _DECAY_MASK = _build_decay_mask()
    return _DECAY_MASK


def get_decay_mask() -> DecayMask:
    return prepare_decay_mask()


class DecayingEntityEffect:
    def __init__(
        self,
        image: pygame.Surface,
        center: tuple[int, int],
        *,
        duration_frames: int = DECAY_EFFECT_DURATION_FRAMES,
        mask: DecayMask | None = None,
    ) -> None:
        self.surface = image.copy()
        self.surface = self.surface.convert_alpha()
        self._apply_grayscale()
        self.rect = self.surface.get_rect(center=center)
        self.duration_frames = max(1, int(duration_frames))
        self.frames_elapsed = 0
        self.mask = mask or get_decay_mask()
        self.mask_index = 0
        self.pixel_carry = 0.0
        self.pixels_per_frame = len(self.mask.positions) / self.duration_frames

    def update(self, *, frames: int = 1) -> bool:
        if frames <= 0:
            return True
        for _ in range(frames):
            if self.frames_elapsed >= self.duration_frames:
                return False
            self.frames_elapsed += 1
            self.pixel_carry += self.pixels_per_frame
            remove_count = int(self.pixel_carry)
            if remove_count <= 0:
                continue
            self.pixel_carry -= remove_count
            self._erase_pixels(remove_count)
        return self.frames_elapsed < self.duration_frames

    def _erase_pixels(self, count: int) -> None:
        width, height = self.surface.get_size()
        for _ in range(count):
            if self.mask_index >= len(self.mask.positions):
                return
            x, y = self.mask.positions[self.mask_index]
            self.mask_index += 1
            if x >= width or y >= height:
                continue
            self.surface.set_at((x, y), (0, 0, 0, 0))

    def _apply_grayscale(self) -> None:
        rgb = pygame.surfarray.pixels3d(self.surface)
        alpha = pygame.surfarray.pixels_alpha(self.surface)
        gray = (0.299 * rgb[:, :, 0] + 0.587 * rgb[:, :, 1] + 0.114 * rgb[:, :, 2]).astype(
            np.uint8
        )
        rgb[:, :, 0] = gray
        rgb[:, :, 1] = gray
        rgb[:, :, 2] = gray
        del rgb, alpha


def update_decay_effects(effects: list[DecayingEntityEffect], *, frames: int = 1) -> None:
    if not effects:
        return
    alive: list[DecayingEntityEffect] = []
    for effect in effects:
        if effect.update(frames=frames):
            alive.append(effect)
    effects[:] = alive


def iter_decay_effects(effects: Iterable[DecayingEntityEffect]) -> Iterable[DecayingEntityEffect]:
    return effects



from __future__ import annotations

import copy
from pathlib import Path
from typing import Any

import pygame
from pygame import surface, time

from ..colors import BLACK, GREEN, LIGHT_GRAY, WHITE
from ..config import DEFAULT_CONFIG, save_config
from ..font_utils import load_font
from ..input_utils import (
    CONTROLLER_BUTTON_DOWN,
    CONTROLLER_BUTTON_DPAD_DOWN,
    CONTROLLER_BUTTON_DPAD_LEFT,
    CONTROLLER_BUTTON_DPAD_RIGHT,
    CONTROLLER_BUTTON_DPAD_UP,
    CONTROLLER_DEVICE_ADDED,
    CONTROLLER_DEVICE_REMOVED,
    init_first_controller,
    init_first_joystick,
    is_confirm_event,
    is_select_event,
)
from ..localization import (
    get_font_settings,
    get_language,
    get_language_name,
    language_options,
    set_language,
)
from ..localization import (
    translate as tr,
)
from ..render import blit_text_wrapped, wrap_text
from ..progress import user_progress_path
from ..screens import TITLE_HEADER_Y, TITLE_SECTION_TOP
from ..windowing import (
    adjust_menu_logical_size,
    nudge_menu_window_scale,
    present,
    sync_window_size,
    toggle_fullscreen,
)


def settings_screen(
    screen: surface.Surface,
    clock: time.Clock,
    config: dict[str, Any],
    fps: int,
    *,
    config_path: Path,
    screen_size: tuple[int, int],
) -> dict[str, Any]:
    """Settings menu shown from the title screen."""

    screen_width, screen_height = screen.get_size()
    if screen_width <= 0 or screen_height <= 0:
        screen_width, screen_height = screen_size
    working = copy.deepcopy(config)
    set_language(working.get("language"))
    selected = 0
    languages = language_options()
    language_codes = [lang.code for lang in languages]
    controller = init_first_controller()
    joystick = init_first_joystick() if controller is None else None

    def _ensure_parent(path: tuple[str, ...]) -> tuple[dict, str]:
        node = working
        for key in path[:-1]:
            node = node.setdefault(key, {})
        return node, path[-1]

    def _get_value(path: tuple[str, ...], default: Any) -> Any:
        node = working
        for key in path[:-1]:
            next_node = node.get(key)
            if not isinstance(next_node, dict):
                return default
            node = next_node
        if isinstance(node, dict):
            return node.get(path[-1], default)
        return default

    def set_value(path: tuple[str, ...], value: Any) -> None:
        node, leaf = _ensure_parent(path)
        node[leaf] = value

    def toggle_row(row: dict) -> None:
        current = bool(_get_value(row["path"], row.get("easy_value", True)))
        set_value(row["path"], not current)

    def set_easy_value(row: dict, use_easy: bool) -> None:
        target = row.get("easy_value", True)
        set_value(row["path"], target if use_easy else not target)

    def cycle_choice(row: dict, direction: int) -> None:
        values = row.get("choices", [])
        if not values:
            return
        current = _get_value(row["path"], values[0])
        try:
            idx = values.index(current)
        except ValueError:
            idx = 0
        idx = (idx + direction) % len(values)
        new_value = values[idx]
        set_value(row["path"], new_value)
        on_change = row.get("on_change")
        if on_change:
            on_change(new_value)

    def build_sections() -> list[dict]:
        return [
            {
                "label": tr("settings.sections.menu"),
                "rows": [
                    {
                        "type": "action",
                        "label": tr("settings.rows.return_to_title"),
                    }
                ],
            },
            {
                "label": tr("settings.sections.localization"),
                "rows": [
                    {
                        "type": "choice",
                        "label": tr("settings.rows.language"),
                        "path": ("language",),
                        "choices": language_codes,
                        "get_display": get_language_name,
                        "on_change": set_language,
                    }
                ],
            },
            {
                "label": tr("settings.sections.player_support"),
                "rows": [
                    {
                        "label": tr("settings.rows.footprints"),
                        "path": ("footprints", "enabled"),
                        "easy_value": True,
                        "left_label": tr("common.on"),
                        "right_label": tr("common.off"),
                    },
                    {
                        "label": tr("settings.rows.car_hint"),
                        "path": ("car_hint", "enabled"),
                        "easy_value": True,
                        "left_label": tr("common.on"),
                        "right_label": tr("common.off"),
                    },
                ],
            },
            {
                "label": tr("settings.sections.tougher_enemies"),
                "rows": [
                    {
                        "label": tr("settings.rows.fast_zombies"),
                        "path": ("fast_zombies", "enabled"),
                        "easy_value": False,
                        "left_label": tr("common.off"),
                        "right_label": tr("common.on"),
                    },
                    {
                        "label": tr("settings.rows.steel_beams"),
                        "path": ("steel_beams", "enabled"),
                        "easy_value": False,
                        "left_label": tr("common.off"),
                        "right_label": tr("common.on"),
                    },
                ],
            },
            {
                "label": tr("settings.sections.visual"),
                "rows": [
                    {
                        "label": tr("settings.rows.shadows"),
                        "path": ("visual", "shadows", "enabled"),
                        "easy_value": True,
                        "left_label": tr("common.on"),
                        "right_label": tr("common.off"),
                    },
                ],
            },
        ]

    def rebuild_rows() -> tuple[list[dict], list[dict], list[str]]:
        current_sections = build_sections()
        flat_rows: list[dict] = []
        flat_sections: list[str] = []
        for section in current_sections:
            for row in section["rows"]:
                flat_rows.append(row)
                flat_sections.append(section["label"])
        return current_sections, flat_rows, flat_sections

    sections, rows, row_sections = rebuild_rows()
    row_count = len(rows)
    last_language = get_language()

    def _exit_settings() -> dict[str, Any]:
        save_config(working, config_path)
        return working

    def _render_frame() -> None:
        nonlocal last_language, sections, rows, row_sections, row_count, selected
        current_language = get_language()
        if current_language != last_language:
            sections, rows, row_sections = rebuild_rows()
            row_count = len(rows)
            selected %= row_count
            last_language = current_language

        screen.fill(BLACK)
        try:
            font_settings = get_font_settings()
            highlight_color = (70, 70, 70)
            title_text = tr("settings.title")
            title_font = load_font(font_settings.resource, font_settings.scaled_size(33))
            title_lines = wrap_text(title_text, title_font, screen_width)
            title_line_height = int(
                round(title_font.get_linesize() * font_settings.line_height_scale)
            )
            title_height = max(1, len(title_lines)) * title_line_height
            title_width = max(
                (title_font.size(line)[0] for line in title_lines if line), default=0
            )
            title_topleft = (
                screen_width // 2 - title_width // 2,
                TITLE_HEADER_Y - title_height // 2,
            )
            blit_text_wrapped(
                screen,
                title_text,
                title_font,
                LIGHT_GRAY,
                title_topleft,
                screen_width,
                line_height_scale=font_settings.line_height_scale,
            )

            row_height = (
                int(
                    round(
                        load_font(
                            font_settings.resource, font_settings.scaled_size(11)
                        ).get_linesize()
                        * font_settings.line_height_scale
                    )
                )
                + 2
            )
            start_y = TITLE_SECTION_TOP

            segment_width = int(round(30 * 1.5 * 0.8))
            segment_height = int(round(18 * 0.8))
            segment_gap = 10
            segment_total_width = segment_width * 2 + segment_gap

            column_margin = 24
            column_width = screen_width // 2 - column_margin * 2
            section_spacing = 4
            row_indent = 12
            value_padding = 20

            section_states: dict[str, dict] = {}
            y_cursor = start_y
            header_font = load_font(font_settings.resource, font_settings.scaled_size(11))
            for section in sections:
                header_lines = wrap_text(section["label"], header_font, column_width)
                header_line_height = int(
                    round(header_font.get_linesize() * font_settings.line_height_scale)
                )
                header_height = max(1, len(header_lines)) * header_line_height
                section_states[section["label"]] = {
                    "next_y": y_cursor + header_height + 4,
                    "header_text": section["label"],
                    "header_pos": (column_margin, y_cursor),
                }
                rows_in_section = len(section["rows"])
                y_cursor = (
                    section_states[section["label"]]["next_y"]
                    + rows_in_section * row_height
                    + section_spacing
                )

            for state in section_states.values():
                blit_text_wrapped(
                    screen,
                    state["header_text"],
                    header_font,
                    LIGHT_GRAY,
                    state["header_pos"],
                    column_width,
                    line_height_scale=font_settings.line_height_scale,
                )

            label_font = load_font(font_settings.resource, font_settings.scaled_size(11))
            value_font = load_font(font_settings.resource, font_settings.scaled_size(11))
            for idx, row in enumerate(rows):
                section_label = row_sections[idx]
                state = section_states[section_label]
                col_x = column_margin + row_indent
                row_width = column_width - row_indent + value_padding
                row_type = row.get("type", "toggle")
                value = None
                if row_type != "action":
                    value = _get_value(
                        row["path"],
                        row.get("easy_value", row.get("choices", [None])[0]),
                    )
                row_y_current = state["next_y"]
                state["next_y"] += row_height

                highlight_rect = pygame.Rect(
                    col_x, row_y_current - 2, row_width, row_height
                )
                if idx == selected:
                    pygame.draw.rect(screen, highlight_color, highlight_rect)

                label_height = int(
                    round(
                        label_font.get_linesize() * font_settings.line_height_scale
                    )
                )
                blit_text_wrapped(
                    screen,
                    row["label"],
                    label_font,
                    WHITE,
                    (
                        col_x,
                        row_y_current + (row_height - label_height) // 2,
                    ),
                    row_width,
                    line_height_scale=font_settings.line_height_scale,
                )
                if row_type == "choice":
                    display_fn = row.get("get_display")
                    display_text = (
                        display_fn(value)
                        if display_fn and value is not None
                        else str(value)
                    )
                    text_width = value_font.size(display_text)[0]
                    text_height = int(
                        round(
                            value_font.get_linesize() * font_settings.line_height_scale
                        )
                    )
                    blit_text_wrapped(
                        screen,
                        display_text,
                        value_font,
                        WHITE,
                        (
                            col_x + row_width - text_width,
                            row_y_current + (row_height - text_height) // 2,
                        ),
                        row_width,
                        line_height_scale=font_settings.line_height_scale,
                    )
                elif row_type == "toggle":
                    slider_y = row_y_current + (row_height - segment_height) // 2 - 2
                    slider_x = col_x + row_width - segment_total_width
                    left_rect = pygame.Rect(
                        slider_x, slider_y, segment_width, segment_height
                    )
                    right_rect = pygame.Rect(
                        left_rect.right + segment_gap,
                        slider_y,
                        segment_width,
                        segment_height,
                    )

                    left_active = value == row["easy_value"]
                    right_active = not left_active

                    def draw_segment(
                        rect: pygame.Rect, text: str, active: bool
                    ) -> None:
                        base_color = (35, 35, 35)
                        active_color = (60, 90, 60) if active else base_color
                        outline_color = GREEN if active else LIGHT_GRAY
                        pygame.draw.rect(screen, active_color, rect)
                        pygame.draw.rect(screen, outline_color, rect, width=2)
                        text_width = value_font.size(text)[0]
                        text_height = int(
                            round(
                                value_font.get_linesize()
                                * font_settings.line_height_scale
                            )
                        )
                        blit_text_wrapped(
                            screen,
                            text,
                            value_font,
                            WHITE,
                            (
                                rect.centerx - text_width // 2,
                                rect.centery - text_height // 2,
                            ),
                            rect.width,
                            line_height_scale=font_settings.line_height_scale,
                        )

                    draw_segment(left_rect, row["left_label"], left_active)
                    draw_segment(right_rect, row["right_label"], right_active)

            hint_start_y = start_y
            hint_start_x = screen_width // 2 + 16
            hint_font = load_font(font_settings.resource, font_settings.scaled_size(11))
            hint_lines = [
                tr("settings.hints.navigate"),
                tr("settings.hints.adjust"),
                tr("settings.hints.toggle"),
                tr("settings.hints.reset"),
                tr("settings.hints.exit"),
            ]
            hint_line_height = int(
                round(hint_font.get_linesize() * font_settings.line_height_scale)
            )
            hint_max_width = screen_width - hint_start_x - 16
            y_cursor = hint_start_y
            for line in hint_lines:
                blit_text_wrapped(
                    screen,
                    line,
                    hint_font,
                    WHITE,
                    (hint_start_x, y_cursor),
                    hint_max_width,
                    line_height_scale=font_settings.line_height_scale,
                )
                y_cursor += hint_line_height

            y_cursor += 26
            window_hint = tr("menu.window_hint")
            window_lines = wrap_text(window_hint, hint_font, hint_max_width)
            window_height = max(1, len(window_lines)) * hint_line_height
            blit_text_wrapped(
                screen,
                window_hint,
                hint_font,
                WHITE,
                (hint_start_x, y_cursor),
                hint_max_width,
                line_height_scale=font_settings.line_height_scale,
            )
            y_cursor += window_height

            config_text = tr("settings.config_path", path=str(config_path))
            progress_text = tr("settings.progress_path", path=str(user_progress_path()))
            path_font = load_font(font_settings.resource, font_settings.scaled_size(11))
            config_width = path_font.size(config_text)[0]
            config_height = int(
                round(path_font.get_linesize() * font_settings.line_height_scale)
            )
            progress_width = path_font.size(progress_text)[0]
            config_top = screen_height - 32 - config_height
            blit_text_wrapped(
                screen,
                config_text,
                path_font,
                LIGHT_GRAY,
                (screen_width // 2 - config_width // 2, config_top),
                screen_width,
                line_height_scale=font_settings.line_height_scale,
            )
            blit_text_wrapped(
                screen,
                progress_text,
                path_font,
                LIGHT_GRAY,
                (screen_width // 2 - progress_width // 2, screen_height - 32),
                screen_width,
                line_height_scale=font_settings.line_height_scale,
            )
        except pygame.error as e:
            print(f"Error rendering settings: {e}")

        present(screen)
        clock.tick(fps)

    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return config
            if event.type in (pygame.WINDOWSIZECHANGED, pygame.VIDEORESIZE):
                sync_window_size(event)
                adjust_menu_logical_size()
                continue
            if event.type == pygame.JOYDEVICEADDED or (
                CONTROLLER_DEVICE_ADDED is not None
                and event.type == CONTROLLER_DEVICE_ADDED
            ):
                if controller is None:
                    controller = init_first_controller()
                if controller is None:
                    joystick = init_first_joystick()
            if event.type == pygame.JOYDEVICEREMOVED or (
                CONTROLLER_DEVICE_REMOVED is not None
                and event.type == CONTROLLER_DEVICE_REMOVED
            ):
                if controller and not controller.get_init():
                    controller = None
                if joystick and not joystick.get_init():
                    joystick = None
            if is_select_event(event):
                return _exit_settings()
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_LEFTBRACKET:
                    nudge_menu_window_scale(0.5)
                    continue
                if event.key == pygame.K_RIGHTBRACKET:
                    nudge_menu_window_scale(2.0)
                    continue
                if event.key == pygame.K_f:
                    toggle_fullscreen()
                    adjust_menu_logical_size()
                    continue
                if event.key in (pygame.K_ESCAPE, pygame.K_BACKSPACE):
                    return _exit_settings()
                if event.key in (pygame.K_UP, pygame.K_w):
                    selected = (selected - 1) % row_count
                if event.key in (pygame.K_DOWN, pygame.K_s):
                    selected = (selected + 1) % row_count
                current_row = rows[selected]
                row_type = current_row.get("type", "toggle")
                if event.key in (pygame.K_SPACE, pygame.K_RETURN):
                    if row_type == "action":
                        return _exit_settings()
                    if row_type == "toggle":
                        toggle_row(current_row)
                    elif row_type == "choice":
                        cycle_choice(current_row, 1)
                if event.key == pygame.K_LEFT and row_type != "action":
                    if row_type == "toggle":
                        set_easy_value(current_row, True)
                    elif row_type == "choice":
                        cycle_choice(current_row, -1)
                if event.key == pygame.K_RIGHT and row_type != "action":
                    if row_type == "toggle":
                        set_easy_value(current_row, False)
                    elif row_type == "choice":
                        cycle_choice(current_row, 1)
                if event.key == pygame.K_r:
                    working = copy.deepcopy(DEFAULT_CONFIG)
                    set_language(working.get("language"))
            if event.type == pygame.JOYBUTTONDOWN or (
                CONTROLLER_BUTTON_DOWN is not None
                and event.type == CONTROLLER_BUTTON_DOWN
            ):
                current_row = rows[selected]
                row_type = current_row.get("type", "toggle")
                if is_confirm_event(event):
                    if row_type == "action":
                        return _exit_settings()
                    if row_type == "toggle":
                        toggle_row(current_row)
                    elif row_type == "choice":
                        cycle_choice(current_row, 1)
                if (
                    CONTROLLER_BUTTON_DOWN is not None
                    and event.type == CONTROLLER_BUTTON_DOWN
                ):
                    if (
                        CONTROLLER_BUTTON_DPAD_UP is not None
                        and event.button == CONTROLLER_BUTTON_DPAD_UP
                    ):
                        selected = (selected - 1) % row_count
                    if (
                        CONTROLLER_BUTTON_DPAD_DOWN is not None
                        and event.button == CONTROLLER_BUTTON_DPAD_DOWN
                    ):
                        selected = (selected + 1) % row_count
                    if (
                        CONTROLLER_BUTTON_DPAD_LEFT is not None
                        and event.button == CONTROLLER_BUTTON_DPAD_LEFT
                        and row_type != "action"
                    ):
                        if row_type == "toggle":
                            set_easy_value(current_row, True)
                        elif row_type == "choice":
                            cycle_choice(current_row, -1)
                    if (
                        CONTROLLER_BUTTON_DPAD_RIGHT is not None
                        and event.button == CONTROLLER_BUTTON_DPAD_RIGHT
                        and row_type != "action"
                    ):
                        if row_type == "toggle":
                            set_easy_value(current_row, False)
                        elif row_type == "choice":
                            cycle_choice(current_row, 1)
            if event.type == pygame.JOYHATMOTION:
                current_row = rows[selected]
                row_type = current_row.get("type", "toggle")
                hat_x, hat_y = event.value
                if hat_y == 1:
                    selected = (selected - 1) % row_count
                elif hat_y == -1:
                    selected = (selected + 1) % row_count
                if hat_x == -1 and row_type != "action":
                    if row_type == "toggle":
                        set_easy_value(current_row, True)
                    elif row_type == "choice":
                        cycle_choice(current_row, -1)
                elif hat_x == 1 and row_type != "action":
                    if row_type == "toggle":
                        set_easy_value(current_row, False)
                    elif row_type == "choice":
                        cycle_choice(current_row, 1)

        _render_frame()
        present(screen)
        clock.tick(fps)
